#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.types.structural.dynamic import CtyDynamic
from pyvider.cty.types.structural.object import CtyObject
from pyvider.cty.types.structural.tuple import CtyTuple

#
# pyvider/cty/types/structural/__init__.py
#
"""
CTY Structural Types.

This package implements the structural types for the CTY system,
including Dynamic (any type), Object (fixed-key map), and Tuple
(fixed-sequence, heterogeneous list).
"""

__all__ = [
    "CtyDynamic",
    "CtyObject",
    "CtyTuple",
]

# 🌊🪢🔚
