#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

"""Configuration and defaults for pyvider-cty."""

# 🌊🪢🔚
