#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.types.capsule import CtyCapsule

"""
Defines standard, built-in capsule types for pyvider.cty.
"""

BytesCapsule = CtyCapsule("Bytes", bytes)
"""A capsule type for wrapping raw bytes."""

# 🌊🪢🔚
